import requests
import argparse
import os
import sys
import pandas as pd
import numpy as np
import bs4
from typing import Union
from datetime import datetime,timedelta
from mappings import PSRTYPE_MAPPINGS, BZN, CTA, CTY, country_dict
from parsers import parse_generation


GENERATION_ELEMENT = "inBiddingZone_Domain.mRID"
CONSUMPTION_ELEMENT = "outBiddingZone_Domain.mRID"


def make_api_request(token, period_start, period_end, in_domain,documentType, prodtype):
    url = 'https://web-api.tp.entsoe.eu/api?documentType='+documentType+'&in_Domain='+in_domain+'&securityToken='+token+'&periodStart='+period_start+'&periodEnd='+period_end+'&processType='+prodtype
    response = requests.get(url)
    print("url",url)
    return response.text


def run_generation_per_plant(token, country_code, period_start, period_end, filename, country):
    """
    Producing a daily output of Entsoe data per plant per country
    """
    capacity_period_start = (datetime.strptime(period_end, "%Y%m%d") - timedelta(days=380)).strftime("%Y%m%d%H%M")
    period_start = (datetime.strptime(period_start, "%Y%m%d")).strftime("%Y%m%d%H%M")
    period_end = (datetime.strptime(period_end, "%Y%m%d") - timedelta(minutes=1)).strftime("%Y%m%d%H%M")
    print(period_start)
    print(period_end)


    documentType = "A73"
    prodtype = "A16"

    text = make_api_request(token, period_start, period_end, country_code,documentType, prodtype)
    df = parse_generation(text, prodtype=prodtype, per_plant=True, include_eic=True)
    #df = parse_generation(text, include_eic=True)
    df["AreaCode"] = country_code
    #Reset index
    df = df.reset_index()
    df = df.rename(columns={"index":"DateTime"})
    

    documentType = "A71"
    prodtype = "A33"

    

    capacity_text = make_api_request(token, capacity_period_start, period_end, country_code,documentType, prodtype)
    capacity_df = parse_generation(capacity_text, prodtype=prodtype, per_plant=True, include_eic=True)
    capacity_df = capacity_df.reset_index()
    #print(capacity_df)
    print(df)

    #Merge dfs
    df = df.merge(capacity_df[["InstalledGenCapacity", "PowerSystemResourceName"]], on ="PowerSystemResourceName", how="left")

    df["MapCode"] = country
    df["AreaTypeCode"] = "CTA"
    df["AreaName"] = "{} CTA".format(country)

    if 'ActualConsumption' not in df.columns:
        df['ActualConsumption'] = np.nan
    elif 'ActualGenerationOutput' not in df.columns:
        df['ActualGenerationOutput'] = np.nan
    elif "InstalledGenCapacity" not in df.columns:
        df["InstalledGenCapacity"] = np.nan

    cols = ["DateTime",	"ResolutionCode", "AreaCode", "AreaTypeCode", "AreaName", "MapCode", "GenerationUnitEIC", "PowerSystemResourceName", "ProductionType", "ActualGenerationOutput", "ActualConsumption", "InstalledGenCapacity"]
    df = df[cols]
   
    
    df = df.sort_values(["DateTime", "ProductionType"])
    if "Unnamed: 0" in df.columns:
        df = df.drop("Unnamed: 0")
    
    return df


def run_generation_aggregated(token, country_code, period_start, period_end, filename, country):
    """
    Producing a daily output of Entsoe data per plant per country
    """
    documentType = "A75"
    prodtype = "A16"

    period_start = (datetime.strptime(period_start, "%Y%m%d")).strftime("%Y%m%d%H%M")
    period_end = (datetime.strptime(period_end, "%Y%m%d")).strftime("%Y%m%d%H%M")
    print(period_start)
    print(period_end)

    cols = ['ResolutionCode', 'AreaCode', 'AreaTypeCode', 'AreaName', 'MapCode', 'ProductionType', 'ActualGenerationOutput', 'ActualConsumption']

    if country in BZN:
        text = make_api_request(token, period_start, period_end, country_code,documentType, prodtype)
    
        bzn_df = parse_generation(text, prodtype=prodtype, per_plant=False, include_eic=False)
        bzn_df["AreaCode"] = country_code
        bzn_df["MapCode"] = country
        bzn_df["AreaTypeCode"] = "BZN"
        bzn_df["AreaName"] = "{} BZN".format(country)
    else:
        bzn_df = pd.DataFrame(columns = cols)

    if country in CTA:
        text = make_api_request(token, period_start, period_end, country_code,documentType, prodtype)
    
        cta_df = parse_generation(text, prodtype=prodtype, per_plant=False, include_eic=False)
        cta_df["AreaCode"] = country_code
        cta_df["MapCode"] = country
        cta_df["AreaTypeCode"] = "CTA"
        cta_df["AreaName"] = "{} CTA".format(country)
    else:
        cta_df = pd.DataFrame(columns = cols)


    if country in CTY:
        text = make_api_request(token, period_start, period_end, country_code,documentType, prodtype)
    
        cty_df = parse_generation(text, prodtype=prodtype, per_plant=False, include_eic=False)
        cty_df["AreaCode"] = country_code
        cty_df["MapCode"] = country
        cty_df["AreaTypeCode"] = "CTY"
        cty_df["AreaName"] = "{} CTY".format(country)
    else:
        cty_df = pd.DataFrame(columns = cols)

    df = pd.concat([bzn_df, cta_df, cty_df])

    if "DateTime" not in df.columns:
        df = df.reset_index()
        df = df.rename(columns={"index":"DateTime"})
        print(df)
    


    if 'ActualConsumption' not in df.columns:
        df['ActualConsumption'] = np.nan
    elif 'ActualGenerationOutput' not in df.columns:
        df['ActualGenerationOutput'] = np.nan

    final_cols = ["DateTime", 'ResolutionCode', 'AreaCode', 'AreaTypeCode', 'AreaName', 'MapCode', 'ProductionType', 'ActualGenerationOutput', 'ActualConsumption']
    df = df[final_cols]

    df = df.sort_values(["DateTime", "ProductionType"])
    if "Unnamed: 0" in df.columns:
        df = df.drop("Unnamed: 0")
    
    return df
    

def getOptions(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-s", "--startdate", help="Initial date in YYYYMMDD format.")
    parser.add_argument("-e", "--enddate", help="End date in YYYYMMDD format.")

    options = parser.parse_args(args)
    return options


def main_16_1_B_C(token, period_start, period_end):
    """
    Run daily AggregatedGenerationPerType
    """
    outdir_bc = "/esarchive/recon/entsoe/16.1.B_C/daily/{}".format(period_start)
    dfs = []
    if not os.path.exists(outdir_bc):
        os.makedirs(outdir_bc)

    # Running for all zones then outputting to combined csv
    filename = "{}/{}_AggregatedGenerationPerType_16.1.B_C.csv".format(outdir_bc, period_start)
    for country, info in country_dict.items():
        print(country)
        print(info["code"])
        country_df = run_generation_aggregated(token, info["code"], period_start, period_end, filename, country)
        dfs.append(country_df)

        if country == "ES":
            spain_df = country_df

    df = pd.concat(dfs)
    
    df.to_csv(filename)

    #outputting just for spain
    spain_filename = "{}/{}-ES_AggregatedGenerationPerType_16.1.B_C.csv".format(outdir_bc, period_start)
    spain_df.to_csv(spain_filename)


def main_16_1_A(token, period_start, period_end):
    """
    Run daily ActualGenerationOutputPerGenerationUnit
    """
    outdir_bc = "/esarchive/recon/entsoe/16.1.A/daily/{}".format(period_start)
    dfs = []
    if not os.path.exists(outdir_bc):
        os.makedirs(outdir_bc)

    # Running for all zones then outputting to combined csv
    filename = "{}/{}_ActualGenerationOutputPerGenerationUnit_16.1.A.csv".format(outdir_bc, period_start)
    for country, info in CTA.items():
        print(country)
        print(info["code"])
        country_df = run_generation_per_plant(token, info["code"], period_start, period_end, filename, country)
        dfs.append(country_df)

        if country == "ES":
            spain_df = country_df

    df = pd.concat(dfs)
    print("Outputting full country zones dataframe to csv")
    df.to_csv(filename)

    #outputting just for spain
    spain_filename = "{}/{}-ES_ActualGenerationOutputPerGenerationUnit_16.1.A.csv".format(outdir_bc, period_start)
    print("Outputting spain dataframe to csv")
    spain_df.to_csv(spain_filename)




if __name__ == "__main__":

    token = "2df5e12b-2d52-483c-8fa8-d308e3dfd769"

    options = getOptions()
    startdate = options.startdate
    enddate = options.enddate
    for date in pd.date_range(startdate, enddate):
        period_start = date.strftime("%Y%m%d")
        period_end = (date + timedelta(days=1)).strftime("%Y%m%d")
        print(period_start)
        print(period_end)

        main_16_1_A(token, period_start, period_end)
        main_16_1_B_C(token, period_start, period_end)
