"""

Parsing the data

"""
import pandas as pd
import numpy as np
import bs4
from typing import Union
from mappings import PSRTYPE_MAPPINGS, BZN, CTA, CTY, country_dict


GENERATION_ELEMENT = "inBiddingZone_Domain.mRID"
CONSUMPTION_ELEMENT = "outBiddingZone_Domain.mRID"

def _extract_timeseries(xml_text):
    """
    Parameters
    ----------
    xml_text : str

    Yields
    -------
    bs4.element.tag
    """
    if not xml_text:
        return
    soup = bs4.BeautifulSoup(xml_text, 'html.parser')
    
    for timeseries in soup.find_all('timeseries'):
        yield timeseries

def parse_generation(
        xml_text: str,
        prodtype: str,
        per_plant: bool = False,
        include_eic: bool = False, 
        nett: bool = False) -> Union[pd.DataFrame, pd.Series]:
    """
    Parameters
    ----------
    xml_text : str
    per_plant : bool
        Decide if you need the parser that can extract plant info as well.
    nett : bool
        If you want to condense generation and consumption of a plant into a
        nett number
    include_eic: bool
        If you want to include the eic code of a plan in the output

    Returns
    -------
    pd.DataFrame 
    """
    dfs = []
    for soup in _extract_timeseries(xml_text):
        df = _parse_generation_timeseries(soup, prodtype, per_plant=per_plant, include_eic=include_eic)
        dfs.append(df)
        
    cols = ["ResolutionCode", "AreaCode", "AreaTypeCode", "AreaName", "MapCode", "GenerationUnitEIC", "PowerSystemResourceName", "ProductionType", "ActualGenerationOutput", "ActualConsumption", "InstalledGenCapacity"]
    if dfs:
        df = pd.concat(dfs)
    else:
        df = pd.DataFrame(columns = cols)
        

    return df

def _parse_generation_timeseries(soup, prodtype, per_plant: bool = False, include_eic: bool = False) -> pd.Series:
    """
    Works for generation by type, generation forecast, and wind and solar
    forecast

    Parameters
    ----------
    soup : bs4.element.tag

    Returns
    -------
    pd.DataFrame
    """
    
    positions = []
    quantities = []
    for point in soup.find_all('point'):
        positions.append(int(point.find('position').text))
        quantity = point.find('quantity')
        if quantity is None:
            raise LookupError('No quantity found in this point, it should have one: {}'.format(point))
        quantities.append(float(quantity.text))

    if soup.find(CONSUMPTION_ELEMENT.lower()):
        metric = 'ActualConsumption'
    else:
        metric = 'ActualGenerationOutput'


    df = pd.DataFrame(index=positions, data=quantities, columns=[metric])
    df = df.sort_index()
    
    
    if len(df) == 1:
        df.index = [soup.find('start').text]
    else:
        df.index = _parse_datetimeindex(soup)

    df["ResolutionCode"] = soup.find('resolution').text


    # Check if there is a psrtype, if so, get it.
    _psrtype = soup.find('psrtype')
    if _psrtype is not None:
        psrtype = _psrtype.text
    else:
        psrtype = None


    # Set both psrtype and metric as columns in the df
    if psrtype:
        psrtype_name = PSRTYPE_MAPPINGS[psrtype]
        df["ProductionType"] = psrtype_name

    if prodtype == "A33":
        key_name = "registeredresource."
        df = df.rename(columns={"ActualGenerationOutput":"InstalledGenCapacity"})
    else:
        key_name = ""

    if per_plant:
        plantname = soup.find('{}name'.format(key_name)).text
        df["PowerSystemResourceName"] = plantname
        if include_eic:
            eic = soup.find("{}mrid".format(key_name), codingscheme="A01").text
            df["GenerationUnitEIC"] = eic


    return df

def _parse_datetimeindex(soup, tz=None):
    """
    Create a datetimeindex from a parsed beautifulsoup,
    given that it contains the elements 'start', 'end'
    and 'resolution'

    Parameters
    ----------
    soup : bs4.element.tag
    tz: str

    Returns
    -------
    pd.DatetimeIndex
    """
    start = pd.Timestamp(soup.find('start').text)
    end = pd.Timestamp(soup.find_all('end')[-1].text)
    if tz is not None:
        start = start.tz_convert(tz)
        end = end.tz_convert(tz)

    delta = _resolution_to_timedelta(res_text=soup.find('resolution').text)
    index = pd.date_range(start=start, end=end, freq=delta, closed='left')
    if tz is not None:
        dst_jump = len(set(index.map(lambda d: d.dst()))) > 1
        if dst_jump and delta == "7D":
            # For a weekly granularity, if we jump over the DST date in October,
            # date_range erronously returns an additional index element
            # because that week contains 169 hours instead of 168.
            index = index[:-1]
        index = index.tz_convert("UTC")

    return index

def _resolution_to_timedelta(res_text: str) -> str:
    """
    Convert an Entsoe resolution to something that pandas can understand
    """
    resolutions = {
        'PT60M': '60min',
        'P1Y': '12M',
        'PT15M': '15min',
        'PT30M': '30min',
        'P1D': '1D',
        'P7D': '7D',
        'P1M': '1M',
    }
    delta = resolutions.get(res_text)
    if delta is None:
        raise NotImplementedError("Sorry, I don't know what to do with the "
                                  "resolution '{}', because there was no "
                                  "documentation to be found of this format. "
                                  "Everything is hard coded. Please open an "
                                  "issue.".format(res_text))
    return delta